#include <ros/ros.h>
#include <pr2_controllers_msgs/Pr2GripperCommandAction.h>
#include <slipgrip_controller/GripperFindContactAction.h>
#include <actionlib/client/simple_action_client.h>

// Our Action interface type, provided as a typedef for convenience
typedef actionlib::SimpleActionClient<slipgrip_controller::GripperFindContactAction> ContactClient;
// Our Action interface type, provided as a typedef for convenience                   
typedef actionlib::SimpleActionClient<pr2_controllers_msgs::Pr2GripperCommandAction> GripperClient;


class Gripper{
private:
  GripperClient* gripper_client_;  
  ContactClient* contact_client_;

public:
  //Action client initialization
  Gripper(){

    //Initialize the client for the Action interface to the gripper controller
    //and tell the action client that we want to spin a thread by default
    gripper_client_ = new GripperClient("r_gripper_fingersensor_controller/gripper_action", true);
    contact_client_  = new ContactClient("r_gripper_fingersensor_controller/find_contact",true);

    //wait for the gripper action server to come up 
    while(!gripper_client_->waitForServer(ros::Duration(5.0))){
      ROS_INFO("Waiting for the r_gripper_fingersensor_controller/gripper_action action server to come up");
    }

    while(!contact_client_->waitForServer(ros::Duration(5.0))){
      ROS_INFO("Waiting for the r_gripper_fingersensor_controller/find_contact action server to come up");
    }
  }

  ~Gripper(){
    delete gripper_client_;
    delete contact_client_;
  }

  //Open the gripper
  void open(){
    pr2_controllers_msgs::Pr2GripperCommandGoal open;
    open.command.position = 0.08;
    open.command.max_effort = -1.0;  // Do not limit effort (negative)
    
    ROS_INFO("Sending open goal");
    gripper_client_->sendGoal(open);
    gripper_client_->waitForResult();
    if(gripper_client_->getState() == actionlib::SimpleClientGoalState::SUCCEEDED)
      ROS_INFO("The gripper opened!");
    else
      ROS_INFO("The gripper failed to open.");
  }

  //Close the gripper
  void close(){
    pr2_controllers_msgs::Pr2GripperCommandGoal squeeze;
    squeeze.command.position = 0.0;
    squeeze.command.max_effort = 30.0;  // Close gently
    
    ROS_INFO("Sending squeeze goal");
    gripper_client_->sendGoal(squeeze);
    gripper_client_->waitForResult();
    if(gripper_client_->getState() == actionlib::SimpleClientGoalState::SUCCEEDED)
      ROS_INFO("The gripper closed!");
    else
      ROS_INFO("The gripper failed to close.");
  }


  //Find a single finger contact and go into position control mode 
  void findOneContact(){
    slipgrip_controller::GripperFindContactGoal findOne;
    findOne.command.contacts_desired = 3;
    findOne.command.close_speed = 0.05;
    findOne.command.max_effort = -1.0;
    findOne.command.force_control_at_goal = false;
    findOne.command.zero_forces = true;
 
    
    ROS_INFO("Sending find 1 contact goal");
    contact_client_->sendGoal(findOne);
    contact_client_->waitForResult();
    if(contact_client_->getState() == actionlib::SimpleClientGoalState::SUCCEEDED)
    {
      ROS_INFO("Contact found. Left: %d, Right: %d", contact_client_->getResult()->data.left_contact, 
	       contact_client_->getResult()->data.right_contact);
      ROS_INFO("Contact force. Left: %f , Right: %f", contact_client_->getResult()->data.left_pad_contact_force, 
	       contact_client_->getResult()->data.right_pad_contact_force);
    }
    else
      ROS_INFO("The gripper did not find a contact.");
  }



  //Find two contacts and go into force control mode
  void findTwoContacts(){
    slipgrip_controller::GripperFindContactGoal findTwo;
    findTwo.command.contacts_desired = 0;
    findTwo.command.close_speed = 0.05;
    findTwo.command.max_effort = -1.0;
    findTwo.command.force_control_at_goal = true;
    findTwo.command.zero_forces = true;
    
    ROS_INFO("Sending find 2 contact goal");
    contact_client_->sendGoal(findTwo);
    contact_client_->waitForResult(ros::Duration(5.0));
    if(contact_client_->getState() == actionlib::SimpleClientGoalState::SUCCEEDED)
    {
    }
    else
      ROS_INFO("The gripper did not find a contact or could not maintain contact force.");
    ROS_INFO("Contact found. Left: %d, Right: %d", contact_client_->getResult()->data.left_contact, 
	       contact_client_->getResult()->data.right_contact);
    ROS_INFO("Contact force. Left: %f , Right: %f, Avg: %f", contact_client_->getResult()->data.left_pad_contact_force, 
	       contact_client_->getResult()->data.right_pad_contact_force, (contact_client_->getResult()->data.left_pad_contact_force+
	       contact_client_->getResult()->data.right_pad_contact_force)/2.0);
  }
};

int main(int argc, char** argv){
  ros::init(argc, argv, "simple_gripper");

  Gripper gripper;

  gripper.open();
  sleep(1.0);
  gripper.findOneContact();
  sleep(2.0);

  gripper.open();
  sleep(1.0);
  gripper.findTwoContacts();


//gripper.close();

  return 0;
}
