#include <ros/ros.h>
#include <pr2_controllers_msgs/Pr2GripperCommandAction.h>
#include <slipgrip_controller/GripperFindContactAction.h>
#include <slipgrip_controller/GripperSlipServoAction.h>
#include <actionlib/client/simple_action_client.h>

typedef actionlib::SimpleActionClient<slipgrip_controller::GripperSlipServoAction> SlipClient;
// Our Action interface type, provided as a typedef for convenience
typedef actionlib::SimpleActionClient<slipgrip_controller::GripperFindContactAction> ContactClient;
// Our Action interface type, provided as a typedef for convenience                   
typedef actionlib::SimpleActionClient<pr2_controllers_msgs::Pr2GripperCommandAction> GripperClient;
// Our Action interface type, provided as a typedef for convenience                   


class Gripper{
private:
  GripperClient* gripper_client_;  
  ContactClient* contact_client_;
  SlipClient* slip_client_;

public:
  //Action client initialization
  Gripper(){

    //Initialize the client for the Action interface to the gripper controller
    //and tell the action client that we want to spin a thread by default
    gripper_client_ = new GripperClient("r_gripper_fingersensor_controller/gripper_action", true);
    contact_client_  = new ContactClient("r_gripper_fingersensor_controller/find_contact",true);
    slip_client_  = new SlipClient("r_gripper_fingersensor_controller/slip_servo",true);


    //wait for the gripper action server to come up 
    while(!gripper_client_->waitForServer(ros::Duration(5.0))){
      ROS_INFO("Waiting for the r_gripper_fingersensor_controller/gripper_action action server to come up");
    }

    while(!contact_client_->waitForServer(ros::Duration(5.0))){
      ROS_INFO("Waiting for the r_gripper_fingersensor_controller/find_contact action server to come up");
    }
    
    while(!slip_client_->waitForServer(ros::Duration(5.0))){
      ROS_INFO("Waiting for the r_gripper_fingersensor_controller/slip_servo action server to come up");
    }    
  }

  ~Gripper(){
    delete gripper_client_;
    delete contact_client_;
    delete slip_client_;

  }

  //Open the gripper
  void open(){
    pr2_controllers_msgs::Pr2GripperCommandGoal open;
    open.command.position = 0.09;    // position open (9 cm)
    open.command.max_effort = -1.0;  // Do not limit effort (negative)
    
    ROS_INFO("Sending open goal");
    gripper_client_->sendGoal(open);
    gripper_client_->waitForResult();
    if(gripper_client_->getState() == actionlib::SimpleClientGoalState::SUCCEEDED)
      ROS_INFO("The gripper opened!");
    else
      ROS_INFO("The gripper failed to open.");
  }




  //Find two contacts and go into force control mode
  void findTwoContacts(){
    slipgrip_controller::GripperFindContactGoal findTwo;
    findTwo.command.contacts_desired = 0;            // close until both figners contact
    findTwo.command.close_speed = 0.05;             // .05 m/s
    findTwo.command.max_effort = -1.0;              // unlimited motor effort
    findTwo.command.force_control_at_goal = true;   // on contact move into force control mode using contact force
    findTwo.command.zero_forces = true;             // zero out the forces before we find contact
 
    ROS_INFO("Sending find 2 contact goal");
    contact_client_->sendGoal(findTwo);
    contact_client_->waitForResult(ros::Duration(5.0));
    if(contact_client_->getState() == actionlib::SimpleClientGoalState::SUCCEEDED)
    {
      ROS_INFO("Contact found. Left: %d, Right: %d", contact_client_->getResult()->data.left_contact, 
	       contact_client_->getResult()->data.right_contact);
      ROS_INFO("Contact force. Left: %f , Right: %f", contact_client_->getResult()->data.left_pad_contact_force, 
	       contact_client_->getResult()->data.right_pad_contact_force);
    }
    else
      ROS_INFO("The gripper did not find a contact.");
  }
  
  
  //Slip servo the robot
  void slipServo(){
    slipgrip_controller::GripperSlipServoGoal slip_goal;
    slip_goal.command.deformation_limit = 0.06;    // 2 cm deformation limit
    slip_goal.command.fingertip_start_force = -1.0; // use default start force (current force)  
    slip_goal.command.max_effort = -1.0;            // unlimited motor effort
    slip_goal.command.fingertip_force_limit = -1.0; // no limit on fingertip force

    ROS_INFO("Slip Servoing");
    slip_client_->sendGoal(slip_goal);
    //slip_client_->waitForResult();   //thre is no reason to wait since this action never returns a result
    if(slip_client_->getState() == actionlib::SimpleClientGoalState::SUCCEEDED)
    {
    }
    else
      ROS_INFO("The gripper did not find a contact or could not maintain contact force.");
    ROS_INFO("Contact found. Left: %d, Right: %d", contact_client_->getResult()->data.left_contact, 
	       contact_client_->getResult()->data.right_contact);
    ROS_INFO("Contact force. Left: %f , Right: %f, Avg: %f", contact_client_->getResult()->data.left_pad_contact_force, 
	       contact_client_->getResult()->data.right_pad_contact_force, (contact_client_->getResult()->data.left_pad_contact_force+
	       contact_client_->getResult()->data.right_pad_contact_force)/2.0);
  }  
};

int main(int argc, char** argv){
  ros::init(argc, argv, "simple_gripper");

  Gripper gripper;


  gripper.open();
  sleep(3.0);
  gripper.findTwoContacts();
  gripper.slipServo();

  return 0;
}
