#include <pr2_controller_interface/controller.h>
#include <slipgrip_controller/gripperAction.h>
#include <slipgrip_controller/pressureObserver.h>
#include <slipgrip_controller/AccelerationSensorData.h>
#include <slipgrip_controller/accelerationObserver.h>
#include <slipgrip_controller/rtstorage.h>
#include <pr2_hardware_interface/hardware_interface.h>
#include <slipgrip_controller/GripperJointState.h>
#include <slipgrip_controller/GripperFindContactData.h>
#include <slipgrip_controller/GripperFindContactCommand.h>
#include <slipgrip_controller/GripperSlipServoData.h>
#include <slipgrip_controller/GripperSlipServoCommand.h>
#include <slipgrip_controller/GripperPlaceData.h>
#include <slipgrip_controller/GripperPlaceCommand.h>
#include <std_srvs/Empty.h>
#include <std_msgs/Empty.h>

#include <ros/ros.h>
#include <slipgrip_controller/SlipGripRequest.h>

#include <pr2_controllers_msgs/JointControllerState.h>
#include <pr2_controllers_msgs/Pr2GripperCommand.h>
#include <boost/scoped_ptr.hpp>
#include <boost/thread/condition.hpp>
#include <realtime_tools/realtime_box.h>
#include <realtime_tools/realtime_publisher.h>
#include <std_msgs/Float64.h>


namespace slipgrip_controller{

class SlipGripControllerClass: public pr2_controller_interface::Controller
{
private:
  // our ROS node handle
  ros::NodeHandle nodeHandle;
  // our robot state handle
  pr2_mechanism_model::RobotState *robotHandle;

  // functions
  bool initializeHandles(pr2_mechanism_model::RobotState *robot, ros::NodeHandle &n);

  // variables ---------------
  int loop_count_;         // loop counter
  int contactCounter;      // counter to delay a certain amount of tmie after contact
  bool stable_contact;     // state variable to indicate whether our contact has been held enough
  bool deformableSqueezed;  
  bool placedState;
  double rampTime;         // the time we want any calls to forceRampTo to run for
  double findContact_delay;
  int placeConditions;
  
  // counter to track the time
  ros::Time last_time_;
  ros::Time contact_start_time;

  // uncomment for debugging data storage
  ///*
  control::RTStorage<slipgrip_controller::AccelerationSensorData> storageAcc_;
  control::RTStorage<slipgrip_controller::GripperSensorData> storagePraw_;
  control::RTStorage<slipgrip_controller::GripperSensorData> storagePfilt_;
  control::RTStorage<slipgrip_controller::GripperJointState> storageJoint_;
  //*/

  // handles to our sensors and joint
  pr2_hardware_interface::Accelerometer* accelerometerHandle;
  AccelerationSensorData accData;

  // accelerometer observer
  accelerationObserver *myAccelerationObserver;

  // gripper action object to control our gripper
  gripperAction *myGripperAction;

  // pressure observer object to do analysis on the incoming pressure sensor data
  pressureObserver *myPressureObserver;
  
  //our service publishers
  //ros::ServiceServer forceServo_srv_;
  //ros::ServiceServer positionServo_srv_;
  //ros::ServiceServer motorsOff_srv_;
  //ros::ServiceServer findContact_srv_;
  ros::ServiceServer updateZeros_srv_;
  //ros::ServiceServer slipServo_srv_;
  //ros::ServiceServer disturbanceServo_srv_;
  //ros::ServiceServer getResponse_srv_;

  //our slipgrip request state variable to keep track of controller state
  SlipGripRequest::Response controller_response;
  SlipGripRequest::Request controller_request;
  // blank request/responses
  std_srvs::Empty::Request empty_req;
  std_srvs::Empty::Response empty_resp;

  //our function definitions for the services we will provide
  bool forceServo();
  bool positionServo();
  bool updateZeros(std_srvs::Empty::Request& req,std_srvs::Empty::Response& resp);
  bool findContact();
  bool slipServo();
  //bool disturbanceReject( );
  void reinitializeValues();
  //bool getResponse(slipgrip_controller::SlipGripRequest::Request& req, slipgrip_controller::SlipGripRequest::Response& resp);

  // position data real-time publisher
  //realtime_tools::RealtimeBox<pr2_controllers_msgs::Pr2GripperCommandConstPtr> command_box_;
  boost::scoped_ptr<
    realtime_tools::RealtimePublisher<
    pr2_controllers_msgs::JointControllerState> > controller_state_publisher_ ;
  ros::Subscriber sub_command_;
  void positionCB(const pr2_controllers_msgs::Pr2GripperCommandConstPtr& msg);

  // find_contact real-time publisher
  //realtime_tools::RealtimeBox<slipgrip_controller::GripperFindContactDataConstPtr> findContact_command_box_;
  boost::scoped_ptr<
    realtime_tools::RealtimePublisher<
    slipgrip_controller::GripperFindContactData> > contact_state_publisher_ ;
  ros::Subscriber sub_findcontact_command_;
  void findContactCB(const slipgrip_controller::GripperFindContactCommandConstPtr& msg);

  // slip_servo real-time publisher
  //realtime_tools::RealtimeBox<slipgrip_controller::GripperSlipServoDataConstPtr> slipServo_command_box_;
  boost::scoped_ptr<
    realtime_tools::RealtimePublisher<
    slipgrip_controller::GripperSlipServoData> > slip_state_publisher_ ;
  ros::Subscriber sub_slipservo_command_;
  void slipServoCB(const slipgrip_controller::GripperSlipServoCommandConstPtr& msg);

  // place publisher and listener
  boost::scoped_ptr<
    realtime_tools::RealtimePublisher<
    slipgrip_controller::GripperPlaceData> > place_state_publisher_ ;
  ros::Subscriber sub_place_command_;
  void placeCB(const slipgrip_controller::GripperPlaceCommandConstPtr& msg);
  
public:
  virtual bool init(pr2_mechanism_model::RobotState *robot, ros::NodeHandle &n);
  virtual void update();
  virtual void stopping();
  virtual void starting() {
    // fill in some default data for our real-time message passer
    //pr2_controllers_msgs::Pr2GripperCommandPtr c(new pr2_controllers_msgs::Pr2GripperCommand);
    //c->position = 0.0;
    //c->max_effort = 0.0;
    //command_box_.set(c);

    //slipgrip_controller::GripperFindContactDataPtr d(new slipgrip_controller::GripperFindContactData);
    //findContact_command_box_.set(d);
    
    //slipgrip_controller::GripperSlipServoDataPtr e(new slipgrip_controller::GripperSlipServoData);
    //slipServo_command_box_.set(e);    
  }



};
} 
