#include "slipgrip_controller/contact_observer.h"
#include <pluginlib/class_list_macros.h>

using namespace control;
using namespace contact_observer;

bool ContactObserverClass::initializeHandles(pr2_mechanism_model::RobotState *robot, ros::NodeHandle &n)
{
  /* get a handle to the hardware interface */
  hardwareInterface = robot->model_->hw_;
  if(!hardwareInterface)
  {
      ROS_ERROR("Perhaps Something wrong with the hardware interface pointer!!!!");
  }

  /* get a handle to our accelerometer */
  std::string accelerometer_name;
  if(!n.getParam("accelerometer_name", accelerometer_name))
  {
      ROS_ERROR("No accelerometer given in namespace: '%s')", n.getNamespace().c_str());
      return false;
  }
  
  accelerometerHandle = hardwareInterface->getAccelerometer(accelerometer_name);
  
  if(!accelerometerHandle)
  {
      ROS_ERROR("ContactObserver could not find sensor named '%s'", accelerometer_name.c_str());
      return false;
  }


  /* get a handle to our left finger pressure sensors */
  std::string leftFinger_pressureSensor_name;
  if(!n.getParam("leftFinger_pressureSensor_name", leftFinger_pressureSensor_name))
  {
        ROS_ERROR("No accelerometer given in namespace: '%s')",	n.getNamespace().c_str());
        return false;
  }

  leftFinger_pressureSensorHandle = hardwareInterface->getPressureSensor(leftFinger_pressureSensor_name);

  if(!leftFinger_pressureSensorHandle)
  {
      ROS_ERROR("ContactObserver could not find sensor named '%s'", leftFinger_pressureSensor_name.c_str());
      return false;
  }   


  /* get a handle to our right finger pressure sensors */
  std::string rightFinger_pressureSensor_name;  
  if(!n.getParam("rightFinger_pressureSensor_name", rightFinger_pressureSensor_name))
  {
      ROS_ERROR("No accelerometer given in namespace: '%s')",   n.getNamespace().c_str());
      return false;
  }

  rightFinger_pressureSensorHandle = hardwareInterface->getPressureSensor(rightFinger_pressureSensor_name);

  if(!rightFinger_pressureSensorHandle)
  {
      ROS_ERROR("ContactObserver could not find sensor named '%s'", rightFinger_pressureSensor_name.c_str());
      return false;
  }



  /* get a handle to our desired joint */
  /*
  std::string joint_name;
  if (!n.getParam("joint_name", joint_name))
  {
    ROS_ERROR("No joint given in namespace: '%s')", n.getNamespace().c_str());
    return false;
  }

  joint_state_ = robot->getJointState(joint_name);
  if (!joint_state_)
  {
    ROS_ERROR("ContactObserver could not find joint named '%s'", joint_name.c_str());
    return false;
  }
  */

  return true;
}


/// Controller initialization in non-realtime
bool ContactObserverClass::init(pr2_mechanism_model::RobotState *robot, ros::NodeHandle &n)
{
  // initialize our data storage arrays
  int pressureFrequency = 1000;    // 1khz rt frequency
  int accFrequency = 3000;         // 3khz frequency
  int recordSeconds = 20;     // record for 10 sec
  storage_.init(n,pressureFrequency*recordSeconds,"/contact_observer/pressureStorage");
  storageAcc_.init(n,accFrequency*recordSeconds,"/contact_observer/accStorage");
  
  // initialize handles to our sensors and joints
  bool success =  this->initializeHandles(robot, n);

  // make sure the accelerometer has a 1.5 khz bandwidth
  accelerometerHandle->command_.bandwidth_ = 6;  

  // make sure the accelerometer has a +/- 2g range
  accelerometerHandle->command_.range_ = 0;

  return success;
}



/// Controller startup in realtime
void ContactObserverClass::starting()
{
  //init_pos_ = joint_state_->position_;
}


/// Controller update loop in realtime
void ContactObserverClass::update()
{
  // initialize our data object that we want to store
  slipgrip_controller::GripperSensorData data;
  slipgrip_controller::AccelerationSensorData dataAcc;

  // copy the x-y-z acceleration values to our data object
  std::vector<geometry_msgs::Vector3> threeAccs = accelerometerHandle->state_.samples_;
  for( uint  i = 0; i < threeAccs.size(); i++ )   // should really be threeAccs.size()
  {
    dataAcc.aX = threeAccs[i].x;
    dataAcc.aY = threeAccs[i].y;
    dataAcc.aZ = threeAccs[i].z;
    dataAcc.mytime = ros::Time::now().toSec();

    storageAcc_.store(dataAcc);
    
  }

  // copy the pressure values to our data object
  std::vector< uint16_t > pressureVector_left = leftFinger_pressureSensorHandle->state_.data_;
  std::vector< uint16_t > pressureVector_right = rightFinger_pressureSensorHandle->state_.data_;
  for( int i = 0; i < 22; i ++)    // should really be .size()
  {
    data.pressure_left[i] = (double)pressureVector_left[i];
  }

  for( int i = 0; i < 22; i++)     // should really be .size()
  {
    data.pressure_right[i] = (double)pressureVector_right[i];
  }

  // store the time
  data.rostime = ros::Time::now().toSec();

  // store the data 
  storage_.store(data);
}


/// Controller stopping in realtime
void ContactObserverClass::stopping()
{
 
}

/// Register controller to pluginlib
PLUGINLIB_REGISTER_CLASS(ContactObserverPlugin, 
                         contact_observer::ContactObserverClass, 
                         pr2_controller_interface::Controller)
