/*********************************************************************
 *
 * Software License Agreement (BSD License)
 *
 *  Copyright (c) 2008, Willow Garage, Inc.
 *  All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above
 *     copyright notice, this list of conditions and the following
 *     disclaimer in the documentation and/or other materials provided
 *     with the distribution.
 *   * Neither the name of the Willow Garage nor the names of its
 *     contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 *  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 *  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 *  ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 *  POSSIBILITY OF SUCH DAMAGE.
 *
 *  \author Joe Romano
 *********************************************************************/
//@author  Joe Romano
//@email   joeromano@gmail.com
//@brief   tactile_sensor_observer.cpp

#include <slipgrip_controller/pr2_gripper_sensor_observer.h>
#include <pluginlib/class_list_macros.h>

using namespace slipgrip_controller;

const int num_tactile_sensors_per_finger_pad = 22;

PR2GripperSensorObserver::PR2GripperSensorObserver()
  : loop_counter_(0)
{}
PR2GripperSensorObserver::~PR2GripperSensorObserver()
{}

bool PR2GripperSensorObserver::initializeHandles(pr2_mechanism_model::RobotState *robot, ros::NodeHandle &n)
{
  /* get a handle to the hardware interface */
  pr2_hardware_interface::HardwareInterface* hardware_interface = robot->model_->hw_;
  if(!hardware_interface)
  {
    ROS_ERROR("Perhaps Something wrong with the hardware interface pointer!!!!");
    return false;
  }

  /* get a handle to our accelerometer */
  std::string accelerometer_name;
  if(!n.getParam("accelerometer_name", accelerometer_name))
  {
    ROS_ERROR("No accelerometer given in namespace: '%s')", n.getNamespace().c_str());
    return false;
  }
  
  accelerometer_handle_ = hardware_interface->getAccelerometer(accelerometer_name);
  
  if(!accelerometer_handle_)
  {
    ROS_ERROR("PR2GripperSensorObserver could not find sensor named '%s'", accelerometer_name.c_str());
    return false;
  }

  acceleration_observer_ = new accelerationObserver(accelerometer_handle_);

  /* get a handle to our left finger pressure sensors */
  std::string left_finger_pressure_sensor_name;
  if(!n.getParam("left_finger_pressure_sensor_name", left_finger_pressure_sensor_name))
  {
    ROS_ERROR("No pressure sensor name given in namespace: '%s')",	n.getNamespace().c_str());
    return false;
  }

  pr2_hardware_interface::PressureSensor* left_finger_pressure_sensor_handle = hardware_interface->getPressureSensor(left_finger_pressure_sensor_name);
  if(!left_finger_pressure_sensor_handle)
  {
    ROS_ERROR("PR2GripperSensorObserver could not find sensor named '%s'", left_finger_pressure_sensor_name.c_str());
    return false;
  }   


  /* get a handle to our right finger pressure sensors */
  std::string right_finger_pressure_sensor_name;  
  if(!n.getParam("right_finger_pressure_sensor_name", right_finger_pressure_sensor_name))
  {
    ROS_ERROR("No pressure sensor name given in namespace: '%s')",   n.getNamespace().c_str());
    return false;
  }

  pr2_hardware_interface::PressureSensor* right_finger_pressure_sensor_handle = hardware_interface->getPressureSensor(right_finger_pressure_sensor_name);
  if(!right_finger_pressure_sensor_handle)
  {
    ROS_ERROR("PR2GripperSensorObserver could not find sensor named '%s'", right_finger_pressure_sensor_name.c_str());
    return false;
  }
  pressure_observer_ = new pressureObserver(left_finger_pressure_sensor_handle, right_finger_pressure_sensor_handle);


  /* get a handle to our desired joint */
  std::string joint_name;
  if (!n.getParam("joint_name", joint_name))
  {
    ROS_ERROR("No joint given in namespace: '%s')", n.getNamespace().c_str());
    return false;
  }
  pr2_mechanism_model::JointState* joint_state_ = robot->getJointState(joint_name);
  if (!joint_state_)
  {
    ROS_ERROR("PR2GripperSensorObserver could not find joint named '%s'", joint_name.c_str());
    return false;
  }


  if(!n.getParam("zeroing_duration", zeroing_duration_))
  {
    ROS_ERROR("No zeroing duration given in namespace: '%s')", n.getNamespace().c_str());
    return false;
  }

  loop_counter_ = 0;
  if(!n.getParam("publishing_counter", publishing_counter_))
  {
    ROS_ERROR("No publishing counter given in namespace: '%s')", n.getNamespace().c_str());
    return false;
  }

  return true;
}


/// Controller initialization in non-realtime
bool PR2GripperSensorObserver::init(pr2_mechanism_model::RobotState *robot, ros::NodeHandle &node_handle)
{

  // store the node and robot handle
  assert(robot);
  node_handle_ = node_handle;
  robot_handle_ = robot;

  // initialize handles to our sensors and joints
  bool success =  this->initializeHandles(robot, node_handle);

  // initially place our observer into normal mode
  mode_ = 0;

  // seed the timer
  last_time_ = robot->getTime();

  // setup zeroing times, variables
  zeroing_duration_ = 0.25;
  zeroing_time_ = robot->getTime();
  zeroing_ = false;

  //publish our services
  update_zeros_srv_ = node_handle.advertiseService("update_zeros", &PR2GripperSensorObserver::updateZeros, this);

  //setup our publisher for tactile information
  rosrt::init();
  slipgrip_controller::PR2GripperSensorData pr2_gripper_sensor_data_msg;
  pr2_gripper_sensor_data_msg.left_finger_pad_forces.resize(num_tactile_sensors_per_finger_pad);
  pr2_gripper_sensor_data_msg.right_finger_pad_forces.resize(num_tactile_sensors_per_finger_pad);
  pr2_gripper_sensor_data_msg.left_finger_pad_forces_filtered.resize(num_tactile_sensors_per_finger_pad);
  pr2_gripper_sensor_data_msg.right_finger_pad_forces_filtered.resize(num_tactile_sensors_per_finger_pad);
  int publisher_buffer_size = 2000;
  tactile_state_publisher_.reset(new rosrt::Publisher<slipgrip_controller::PR2GripperSensorData>(node_handle_.advertise<slipgrip_controller::PR2GripperSensorData> (std::string("tactile_sensor_info"), 1), publisher_buffer_size, pr2_gripper_sensor_data_msg));

  return success;
}

/// Controller update loop in realtime
void PR2GripperSensorObserver::update()
{
  // update our timing information
  ros::Time time = robot_handle_->getTime();
  ros::Duration dt = time - last_time_;

  // do some sanity checks
  assert(robot_handle_ != NULL);

  // control modes: 
  //  0 - normal 
  //  1 - rezero  (average the data and apply zero offset in real time until leaving this mode)
  //disabled (do nothing) mode
  if(mode_ == 1)  // update zeros mode
  {
    if(!zeroing_)
    {
      zeroing_time_ = time;
      zeroing_ = true;
    }
    if(zeroing_ && (zeroing_time_ - time >= ros::Duration(zeroing_duration_)))
    {
      zeroing_ = false;
      mode_ = 0;
    }
    if(zeroing_)
      pressure_observer_->updateZeros2();
  }

  // publish tactile sensor information
  publish();

  // call our pressure sensor and accelerometer update algorithm
  pressure_observer_->spin();
  acceleration_observer_->spin();
  
  // store the time
  last_time_ = time;
}

void PR2GripperSensorObserver::publish()
{

  if((loop_counter_ % publishing_counter_) == 0)
    {
      
      slipgrip_controller::PR2GripperSensorDataPtr tactile_state_msg = tactile_state_publisher_->allocate();
      if(tactile_state_msg)
	{
	  tactile_state_msg->header.stamp = ros::Time::now();
	  
	  //boolean indicating whether contact is present
	  tactile_state_msg->left_contact  = pressure_observer_->left_contact;
	  tactile_state_msg->right_contact = pressure_observer_->right_contact;
	  
	  // corrected finger pad forces after zeroing (subtracting zero contact force)
	  tactile_state_msg->left_finger_pad_force = pressure_observer_->padForce_left_cur_nonbiased;
	  tactile_state_msg->right_finger_pad_force = pressure_observer_->padForce_right_cur_nonbiased;
	  // filtered values : high pass filter at 5 Hz after correcting for zero contact
	  tactile_state_msg->left_finger_pad_force_filtered  = pressure_observer_->padForce_left_cur;
	  tactile_state_msg->right_finger_pad_force_filtered = pressure_observer_->padForce_right_cur;
	  
	  // raw data - accelerometer
	  tactile_state_msg->acc_x_raw = acc_data_.aX;
	  tactile_state_msg->acc_y_raw = acc_data_.aY;
	  tactile_state_msg->acc_z_raw = acc_data_.aZ;
	  // filtered data - accelerometer
	  tactile_state_msg->acc_x_filtered = acceleration_observer_->aX;
	  tactile_state_msg->acc_y_filtered = acceleration_observer_->aY;
	  tactile_state_msg->acc_z_filtered = acceleration_observer_->aZ;
	  
	  for(int i=0; i<num_tactile_sensors_per_finger_pad; ++i)
	    {
	      tactile_state_msg->left_finger_pad_forces[i] = pressure_observer_->pressure_current_zerod.pressure_left[i];
	      tactile_state_msg->right_finger_pad_forces[i] = pressure_observer_->pressure_current_zerod.pressure_right[i];
	      
	      tactile_state_msg->left_finger_pad_forces_filtered[i] = pressure_observer_->pressure_cur_bias.pressure_left[i];
	      tactile_state_msg->right_finger_pad_forces_filtered[i] = pressure_observer_->pressure_cur_bias.pressure_right[i];
	    }
	  
	  // publish
	  tactile_state_publisher_->publish(tactile_state_msg);
	}
      else
	{
	  // Skipped to publish a message
	  ROS_WARN("Skipped publishing tactile states message.");
	}  
      
    }
  loop_counter_++;

}

// call to start averaging pressure data to rezero sensors
bool PR2GripperSensorObserver::updateZeros(std_srvs::Empty::Request& req,std_srvs::Empty::Response& resp)
{
  // change the controller state
  mode_ = 1;
  return true;
}

/// Controller stopping in realtime
void PR2GripperSensorObserver::stopping()
{
}

void PR2GripperSensorObserver::reinitializeValues()
{
  // reinialize the grasp contact states
  // pressure_observer_->left_contact = false;
  // pressure_observer_->right_contact = false;
  pressure_observer_->placeContact = false;
  acceleration_observer_->placeContact = false;  
}

/// Register controller to pluginlib
PLUGINLIB_REGISTER_CLASS(PR2GripperSensorObserver, 
                         slipgrip_controller::PR2GripperSensorObserver, 
                         pr2_controller_interface::Controller)

//PLUGINLIB_DECLARE_CLASS(slipgrip_controller, 
//                        PR2GripperSensorObserver, 
//                        slipgrip_controller::PR2GripperSensorObserver, 
//                        pr2_controller_interface::Controller)
