//@author  Joe Romano
//@email   joeromano@gmail.com
//@brief   slipgrip_controller.cpp - controller to do tactile-responsive grasping of objects

#include <slipgrip_controller/slipgrip_controller.h>
#include <pluginlib/class_list_macros.h>
#include <slipgrip_controller/gripperAction.h>
#include <slipgrip_controller/pressureObserver.h>
#include <pr2_hardware_interface/hardware_interface.h>
#include <slipgrip_controller/rtstorage.h>
#include <std_srvs/Empty.h>


using namespace slipgrip_controller;

bool SlipGripControllerClass::initializeHandles(pr2_mechanism_model::RobotState *robot, ros::NodeHandle &n)
{

  /* get a handle to the hardware interface */
  pr2_hardware_interface::HardwareInterface* hardwareInterface = robot->model_->hw_;
  if(!hardwareInterface)
  {
      ROS_ERROR("Perhaps Something wrong with the hardware interface pointer!!!!");
  }

  /* get a handle to our accelerometer */
  std::string accelerometer_name;
  if(!n.getParam("accelerometer_name", accelerometer_name))
  {
      ROS_ERROR("No accelerometer given in namespace: '%s')", n.getNamespace().c_str());
      return false;
  }
  
  accelerometerHandle = hardwareInterface->getAccelerometer(accelerometer_name);
  
  if(!accelerometerHandle)
  {
      ROS_ERROR("SlipGripController could not find sensor named '%s'", accelerometer_name.c_str());
      return false;
  }

  myAccelerationObserver = new accelerationObserver(accelerometerHandle);

  /* get a handle to our left finger pressure sensors */
  std::string leftFinger_pressureSensor_name;
  if(!n.getParam("leftFinger_pressureSensor_name", leftFinger_pressureSensor_name))
  {
        ROS_ERROR("No accelerometer given in namespace: '%s')",	n.getNamespace().c_str());
        return false;
  }

  pr2_hardware_interface::PressureSensor* leftFinger_pressureSensorHandle = hardwareInterface->getPressureSensor(leftFinger_pressureSensor_name);
  if(!leftFinger_pressureSensorHandle)
  {
      ROS_ERROR("SlipGripController could not find sensor named '%s'", leftFinger_pressureSensor_name.c_str());
      return false;
  }   


  /* get a handle to our right finger pressure sensors */
  std::string rightFinger_pressureSensor_name;  
  if(!n.getParam("rightFinger_pressureSensor_name", rightFinger_pressureSensor_name))
  {
      ROS_ERROR("No accelerometer given in namespace: '%s')",   n.getNamespace().c_str());
      return false;
  }

  pr2_hardware_interface::PressureSensor* rightFinger_pressureSensorHandle = hardwareInterface->getPressureSensor(rightFinger_pressureSensor_name);
  if(!rightFinger_pressureSensorHandle)
  {
      ROS_ERROR("SlipGripController could not find sensor named '%s'", rightFinger_pressureSensor_name.c_str());
      return false;
  }
  myPressureObserver = new pressureObserver(leftFinger_pressureSensorHandle, rightFinger_pressureSensorHandle);


  /* get a handle to our desired joint */
  std::string joint_name;
  if (!n.getParam("joint_name", joint_name))
  {
    ROS_ERROR("No joint given in namespace: '%s')", n.getNamespace().c_str());
    return false;
  }
  pr2_mechanism_model::JointState* joint_state_ = robot->getJointState(joint_name);
  if (!joint_state_)
  {
    ROS_ERROR("SlipGripController could not find joint named '%s'", joint_name.c_str());
    return false;
  }
  // instantiate our gripper action object and pass it our joint_state_ reference
  myGripperAction = new gripperAction(joint_state_, myPressureObserver, myAccelerationObserver);


  return true;
}


/// Controller initialization in non-realtime
bool SlipGripControllerClass::init(pr2_mechanism_model::RobotState *robot, ros::NodeHandle &n)
{
  // uncomment for debug data storage
  //*
      // get the name of this controller from our yaml file
      std::string controller_name;
      if(!n.getParam("controller_name", controller_name))
      {
        ROS_ERROR("Could not retrieve controller name from param server: '%s')",   n.getNamespace().c_str());
        return false;
      }

      // initialize our storage vectors
      int accFrequency = 3000;         // 3khz frequency
      int recordSeconds = 30;     // record for 30 sec
      int  pressureFrequency = 1000;         // 3khz frequency                                                                                                                                                                                                    
      storagePraw_.init(n,pressureFrequency*recordSeconds,"/"+controller_name+"/PrawStorage");
      storagePfilt_.init(n,pressureFrequency*recordSeconds,"/"+controller_name+"/PfiltStorage");
      storageJoint_.init(n,pressureFrequency*recordSeconds,"/"+controller_name+"/JointStorage");
      storageAcc_.init(n,accFrequency*recordSeconds,"/"+controller_name+"/accStorage");
  //*/

  // store the node and robot handle
  assert(robot);
  nodeHandle = n;
  robotHandle = robot;

  // initialize handles to our sensors and joints
  bool success =  this->initializeHandles(robot, n);
  loop_count_ = 0;

  // initially place our controller into disabled mode
  controller_response.control_mode = 0;


  // seed the timer
  last_time_ = robot->getTime();

  //publish our services
  //forceServo_srv_ = n.advertiseService("force_servo", &SlipGripControllerClass::forceServo, this);
  updateZeros_srv_ = n.advertiseService("update_zeros", &SlipGripControllerClass::updateZeros, this);
  //disturbanceServo_srv_ = n.advertiseService("disturbance_servo", &SlipGripControllerClass::disturbanceReject, this);
  //getResponse_srv_ = n.advertiseService("get_response", &SlipGripControllerClass::getResponse, this);

  //setup our publisher for joint and motor effort state information
  controller_state_publisher_.reset(new realtime_tools::RealtimePublisher<pr2_controllers_msgs::JointControllerState>(n, "state", 1));

  //setup our publisher for contact state information
  contact_state_publisher_.reset(new realtime_tools::RealtimePublisher<slipgrip_controller::GripperFindContactData>(n, "contact_state", 1));

  //setup our publisher for place state information
  place_state_publisher_.reset(new realtime_tools::RealtimePublisher<slipgrip_controller::GripperPlaceData>(n, "place_state", 1));
  
  //setup our publisher for slip state information
  slip_state_publisher_.reset(new realtime_tools::RealtimePublisher<slipgrip_controller::GripperSlipServoData>(n, "slip_servo_state", 1));

  // setup our callback to handle position command requests
  sub_command_ = n.subscribe<pr2_controllers_msgs::Pr2GripperCommand>("command", 1, &SlipGripControllerClass::positionCB, this);

  // setup our callback to handle find_contact command requests
  sub_findcontact_command_ = n.subscribe<slipgrip_controller::GripperFindContactCommand>("find_contact", 1, &SlipGripControllerClass::findContactCB, this);

  // setup our callback to handle find_contact command requests
  sub_place_command_ = n.subscribe<slipgrip_controller::GripperPlaceCommand>("place", 1, &SlipGripControllerClass::placeCB, this);
  
  // setup our callback to handle slip_servo command requests
  sub_slipservo_command_ = n.subscribe<slipgrip_controller::GripperSlipServoCommand>("slip_servo", 1, &SlipGripControllerClass::slipServoCB, this);

  return success;

}



/// Controller update loop in realtime
void SlipGripControllerClass::update()
{

  // update our timing information
  ros::Time time = robotHandle->getTime();
  ros::Duration dt = time - last_time_;

  // do some sanity checks
  assert(robotHandle != NULL);

  // uncomment for debug data storage
  ///*
  // copy the x-y-z acceleration values to our data object
  std::vector<geometry_msgs::Vector3> threeAccs = accelerometerHandle->state_.samples_;
  for( uint  i = 0; i < threeAccs.size(); i++ )   // should really be threeAccs.size()
  {
    accData.aX = threeAccs[i].x;
    accData.aY = threeAccs[i].y;
    accData.aZ = threeAccs[i].z;
    accData.mytime = time.toSec();

    storageAcc_.store(accData); 
  }

  // store pressure data
  myPressureObserver->pressure_cur_bias.rostime = time.toSec();
  myPressureObserver->pressure_now.rostime = time.toSec();
  storagePfilt_.store(myPressureObserver->pressure_cur_bias);
  storagePraw_.store(myPressureObserver->pressure_now);
  // store joint data
  myGripperAction->gripperState_now.mytime = time.toSec();
  storageJoint_.store(myGripperAction->gripperState_now);
  //*/




  // control modes: 
  //  0 - disabled (command zero effort to motors)
  //  1 - rezero  (average the data and apply zero offset in real time until leaving this mode)
  //  2 - update thresholds (check the current pressure data and update thresholds appropriately) 
  //  3 - position control mode
  //  4 - force control mode
  //  5 - find contact mode (either hold force on contact or place into object-pickup mode)
  //  6 - object pickup mode (increasing slip force)
  //  7 - object place mode

  //disabled (do nothing) mode
  if(controller_response.control_mode == 0){}
  

  // update zeros mode
  else if(controller_response.control_mode == 1)
    myPressureObserver->updateZeros2();

  // position control mode
  else if(controller_response.control_mode == 3)
  {
    // use a zero velocity goal in our position controller to have the robot 
    // gently decelerate when the position error gets small (could be current vel. instead)
    myGripperAction->positionServo(controller_request.servo_position,0.0);
  }

  // force control mode
  else if(controller_response.control_mode == 4)
    myGripperAction->forceServo2(controller_request.servo_force);

  // find contact mode
  else if(controller_response.control_mode == 5)
  { 
    // if we have not yet met the contact requirements
    if(!myGripperAction->grabObject(controller_request.servo_velocity, controller_request.contacts_desired))
    {
      controller_response.contact_success = false;
      // make sure our velocity integral term is starting at our current position in case of contact
      myGripperAction->vel_integral = myGripperAction->gripperState_now.position_;

      // seed our servo force with SOME value
      myGripperAction->servoForce = myGripperAction->forceLightest;
      //ROS_INFO("contact constraints not satisfied!");
    }
    // else if we have met the contact requirements
    else   
    {
	
      // for some time after contact
      if(contactCounter < 50)
      {
	controller_request.servo_position = myGripperAction->positionContact;
	myGripperAction->positionMarker = myGripperAction->positionCurrent;

	// keep track of the highest contact force we've applied
	controller_request.servo_force = myGripperAction->forceContact;	  

	// if we are more than some gap hold our position
	if(myGripperAction->positionCurrent > myGripperAction->positionDropped)
	  myGripperAction->positionServo(controller_request.servo_position,0.0);
	else
	{
	  //ROS_INFO("Pre-estimate DROPPED");
	  myGripperAction->jointState->commanded_effort_ = -100.0; // close as hard as possible
	}
	contact_start_time = robotHandle->getTime();
      } 
      else if(contactCounter == 50)
      {
	controller_request.servo_force = 0.027*(controller_request.servo_force/-controller_request.servo_velocity);
	if( -controller_request.servo_force < -(myGripperAction->forceLightest))
	  controller_request.servo_force = myGripperAction->forceLightest;
      }
      // after our half second pause move into either position or force control
      else if(controller_request.force_control_at_goal) 
      {
	if(myGripperAction->positionCurrent > myGripperAction->positionDropped)
	{
	  //if we have reached some epsilon of our target force then return
	  // or if it took more than 8 seconds to servo our force
	  if( ( (myGripperAction->forceServo2(controller_request.servo_force)) &&
		fabs(myGripperAction->lpVelocity) < 0.001))
	    //  || robotHandle->getTime() - contact_start_time > ros::Duration(8.0))
	  {
	    if(findContact_delay > 1000)
	    {
	      forceServo();
	      stable_contact = true;
	      ROS_INFO("Completed find contact with servo force: %f",myGripperAction->servoForce);
	    }
	    findContact_delay++;
	  }
	  // check the object is deforming a lot (which means it is probably squishy)
	  // NOTE: -.09 should never occur (this is the whole gripper). This code can be activated by lowering making it around -.03
	  else if( myGripperAction->positionCurrent - myGripperAction->positionMarker < -0.09 &&
		   !deformableSqueezed)
	  {
	    controller_request.servo_force = (controller_request.servo_force)*2.0;
	    myGripperAction->positionMarker = myGripperAction->positionCurrent;
	    deformableSqueezed = true;
	    ROS_INFO("Squish!");
	  }
	 
	}
	// if we are less than a gap close the gripper instead of force control
	else
	{
	  //ROS_INFO("FORCE_ESTIMATE DROPPED!");
	  myGripperAction->jointState->commanded_effort_ = -100.0; // close as hard as possible
	  stable_contact = true;
        }
      }
      else
      {
        positionServo();
	stable_contact = true;
      }

      contactCounter++;
      controller_response.contact_success = true;
    }
  }

  // if in slip servo control mode
  else if(controller_response.control_mode == 6)
      myGripperAction->slipServo2();


  // if in place mode
  else if(controller_response.control_mode == 7)
  {
    // if we have not yet met the place conditions hold the current force
    if(!myGripperAction->place(placeConditions))
      myGripperAction->forceServo2(myGripperAction->servoForce);
    else
    {
      if(!myGripperAction->forceRampTo(0.0,rampTime) ){}
      else
      {
	controller_request.servo_position = myGripperAction->positionOpen;
	placedState = true;
	positionServo();
      }
    }
  }


  // else somehow we ended up with some other value for mode (should not be possible)
  else
  {
    ROS_INFO("Slip Grip Controller say WHAAAAAAAT");
    myGripperAction->jointState->commanded_effort_ = 0;
  }    


  // publish information every 10th cycle
  if(loop_count_ % 10 == 0)
  {
      if(controller_state_publisher_ && controller_state_publisher_->trylock())
      {
        controller_state_publisher_->msg_.header.stamp = time;
        controller_state_publisher_->msg_.set_point = controller_request.servo_position;
        controller_state_publisher_->msg_.process_value = myGripperAction->gripperState_now.position_;
        controller_state_publisher_->msg_.process_value_dot = myGripperAction->gripperState_now.velocity_;
        controller_state_publisher_->msg_.error = myGripperAction->gripperState_now.position_ - controller_request.servo_position;
        controller_state_publisher_->msg_.time_step = dt.toSec();
        controller_state_publisher_->msg_.command = myGripperAction->gripperState_now.measured_effort_;

        controller_state_publisher_->unlockAndPublish();
      }

      if(contact_state_publisher_ && contact_state_publisher_->trylock())
      {
        contact_state_publisher_->msg_.header.stamp = time;
        contact_state_publisher_->msg_.left_contact = myPressureObserver->left_contact;
        contact_state_publisher_->msg_.right_contact = myPressureObserver->right_contact;
        contact_state_publisher_->msg_.left_pad_force = myPressureObserver->padForce_left_cur_nonbiased;
        contact_state_publisher_->msg_.right_pad_force = myPressureObserver->padForce_right_cur_nonbiased;
        contact_state_publisher_->msg_.left_pad_contact_force = -(myGripperAction->forceContact_l);
        contact_state_publisher_->msg_.right_pad_contact_force = -(myGripperAction->forceContact_r);
        contact_state_publisher_->msg_.contact_position = myGripperAction->positionContact;
        contact_state_publisher_->msg_.motor_effort = myGripperAction->gripperState_now.measured_effort_;
        contact_state_publisher_->msg_.current_position = myGripperAction->gripperState_now.position_;
	contact_state_publisher_->msg_.stable_contact = stable_contact;
	contact_state_publisher_->msg_.realtime_controller_state = controller_response.control_mode;

        contact_state_publisher_->unlockAndPublish();
      }

      if(slip_state_publisher_ && slip_state_publisher_->trylock())
      {
        slip_state_publisher_->msg_.header.stamp = time;
        slip_state_publisher_->msg_.motor_effort = myGripperAction->gripperState_now.measured_effort_;
        slip_state_publisher_->msg_.fingerpad_force = myPressureObserver->padForce_cur_nonbiased;
        slip_state_publisher_->msg_.deformation = myGripperAction->positionCurrent - myGripperAction->positionMarker;
        slip_state_publisher_->msg_.slip = myGripperAction->slip_flag;
        slip_state_publisher_->msg_.force_limit = myGripperAction->force_limit_flag;
        slip_state_publisher_->msg_.deformation_limit = myGripperAction->deformation_limit_flag;
        slip_state_publisher_->msg_.dropped = myGripperAction->dropped_flag;
	slip_state_publisher_->msg_.realtime_controller_state = controller_response.control_mode;
        
        slip_state_publisher_->unlockAndPublish();
      }

      if(place_state_publisher_ && place_state_publisher_->trylock())
      {
        place_state_publisher_->msg_.header.stamp = time;
        place_state_publisher_->msg_.motor_effort = myGripperAction->gripperState_now.measured_effort_;
        place_state_publisher_->msg_.fingerpad_force = myPressureObserver->padForce_cur_nonbiased;
	place_state_publisher_->msg_.placed = placedState;
	place_state_publisher_->msg_.realtime_controller_state = controller_response.control_mode;
        
        place_state_publisher_->unlockAndPublish();
      }

  }

  // update our gripper action, as well as our accelerometer and pressure states with it
  myGripperAction->updateData();
  
  // increment our counter and store the time
  loop_count_++;
  last_time_ = time;
}


// call to put the robot into servo'd slip mode
bool SlipGripControllerClass::slipServo()
{
  // change the controller state
  controller_response.control_mode = 6;

  ROS_INFO("Starting Slip Servo with: %f N", myGripperAction->servoForce);
  return true;
}


// call to close gripper and look for object contact. once contact is found gripper automatically stops and enters position or force control mode
bool SlipGripControllerClass::findContact()
{
  //ROS_INFO("find contact called");

  // reinialize the grasp contact states
  myPressureObserver->left_contact = false;
  myPressureObserver->right_contact = false;

  // change the controller state
  controller_response.control_mode = 5;

  return true;
}


// call to start averaging pressure data to rezero sensors
bool SlipGripControllerClass::updateZeros(std_srvs::Empty::Request& req,std_srvs::Empty::Response& resp)
{
  // change the controller state
  controller_response.control_mode = 1;
  return true;
}


// call to move into force servo mode
bool SlipGripControllerClass::forceServo()
{
  // only allow negative (closing) forces
  if(controller_request.servo_force <= 0)
  {
    // change the controller state
    controller_response.control_mode = 4;
    // reset the integral term of our force controller to the current position value
    myGripperAction->vel_integral = myGripperAction->gripperState_now.position_;
  }
  else
    ROS_INFO("SlipGrip Controller tried to command positive (gripper opening) servo force. NOT ALLOWED. Refusing to enter force servo mode");

  return true;
}



// call to move into position servo mode
bool SlipGripControllerClass::positionServo()
{
  // change the controller state
  controller_response.control_mode = 3;
  
  ROS_INFO("Gripper Position Servo to: %f", controller_request.servo_position); 

  return true;
}




/// Controller stopping in realtime
void SlipGripControllerClass::stopping()
{
    // turn the motors off before stopping
    myGripperAction->jointState->commanded_effort_ = 0;     
}

void SlipGripControllerClass::reinitializeValues()
{
  // reinitialize force ramp state
  myGripperAction->firstRamp = true;
  deformableSqueezed = false;

  myGripperAction->vel_integral_vcontrol = myGripperAction->positionCurrent;

  // reinialize the grasp contact states
  //myPressureObserver->left_contact = false;
  //myPressureObserver->right_contact = false;
  stable_contact = false;
  placedState = false;
  myPressureObserver->placeContact = false;
  myAccelerationObserver->placeContact = false;

  // reinitialize the contact forces
  myGripperAction->forceContact = 0;
  myGripperAction->forceContact_l = 0;
  myGripperAction->forceContact_r = 0;
  
    // set our fingertip force limit to "unlimited"
  myGripperAction->fingertip_force_limit = 1.0;
  // set our defromation limit to "unlimited"
  myGripperAction->positionMarker_limit = -1.0;
  
  // reset several important variables
  myGripperAction->positionMarker = myGripperAction->positionCurrent;
  myGripperAction->slip_flag = false;
  myGripperAction->force_limit_flag = false;
  myGripperAction->deformation_limit_flag = false;

  // if statement is a  hack for the case where we are closed because the object was dropped
  myGripperAction->dropped_flag = false;
  
  findContact_delay = 0;
}


void SlipGripControllerClass::positionCB(const pr2_controllers_msgs::Pr2GripperCommandConstPtr& msg)
{
  reinitializeValues();
  
  // store the message data
  controller_request.servo_position = msg->position;
  controller_request.max_effort = msg->max_effort;
  myGripperAction->max_effort = msg->max_effort;

  //command_box_.set(msg);
  
  positionServo();
}

void SlipGripControllerClass::findContactCB(const slipgrip_controller::GripperFindContactCommandConstPtr& msg)
{ 
  reinitializeValues();
  contactCounter = 0;
  
  // copy over the data from our message
  controller_request.force_control_at_goal = msg->force_control_at_goal;
  controller_request.servo_velocity = -(msg->close_speed);
  controller_request.max_effort = msg->max_effort;
  myGripperAction->max_effort = msg->max_effort;
  controller_request.contacts_desired = msg->contacts_desired;

  findContact();
}


void SlipGripControllerClass::placeCB(const slipgrip_controller::GripperPlaceCommandConstPtr& msg)
{ 
  reinitializeValues();

  rampTime = msg->release_time;
  placeConditions = msg->contact_conditions;
  controller_response.control_mode = 7;  
}

void SlipGripControllerClass::slipServoCB(const slipgrip_controller::GripperSlipServoCommandConstPtr& msg)
{
  reinitializeValues();

  //set some state variables to be used during slip servoing
  controller_request.deformation_limit = msg->deformation_limit;
  myGripperAction->positionMarker_limit = msg->deformation_limit;
  controller_request.fingertip_start_force = -(msg->fingertip_start_force);

  // check if the user wanted to use the default force or input a force level
  if(msg->fingertip_start_force >= 0.0)
    myGripperAction->servoForce = -(msg->fingertip_start_force);
  else
    myGripperAction->servoForce = controller_request.servo_force;
  controller_request.fingertip_force_limit = -(msg->fingertip_force_limit);
  myGripperAction->fingertip_force_limit = -(msg->fingertip_force_limit);
  controller_request.max_effort = msg->max_effort;
  myGripperAction->max_effort = msg->max_effort;
  
  slipServo();
}


/// Register controller to pluginlib
PLUGINLIB_DECLARE_CLASS(slipgrip_controller,SlipGripControllerPlugin, 
                         slipgrip_controller::SlipGripControllerClass, 
                         pr2_controller_interface::Controller)
